{
  "cells": [
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "---\n",
        "title: \"LSE DS202W 2025: Week 4 - Student Notebook\"\n",
        "subtitle: \"2024/25 Winter Term\"\n",
        "author: \"The DS202 Team\"\n",
        "date: 10 February 2025\n",
        "---\n",
        "\n",
        "# ⚙️ Setup\n",
        "\n",
        "**Import required libraries:**"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 10,
      "metadata": {},
      "outputs": [
        {
          "data": {
            "text/html": [
              "\n",
              "            <div id=\"lZRXqH\"></div>\n",
              "            <script type=\"text/javascript\" data-lets-plot-script=\"library\">\n",
              "                if(!window.letsPlotCallQueue) {\n",
              "                    window.letsPlotCallQueue = [];\n",
              "                }; \n",
              "                window.letsPlotCall = function(f) {\n",
              "                    window.letsPlotCallQueue.push(f);\n",
              "                };\n",
              "                (function() {\n",
              "                    var script = document.createElement(\"script\");\n",
              "                    script.type = \"text/javascript\";\n",
              "                    script.src = \"https://cdn.jsdelivr.net/gh/JetBrains/lets-plot@v4.5.2/js-package/distr/lets-plot.min.js\";\n",
              "                    script.onload = function() {\n",
              "                        window.letsPlotCall = function(f) {f();};\n",
              "                        window.letsPlotCallQueue.forEach(function(f) {f();});\n",
              "                        window.letsPlotCallQueue = [];\n",
              "                        \n",
              "                    };\n",
              "                    script.onerror = function(event) {\n",
              "                        window.letsPlotCall = function(f) {};    // noop\n",
              "                        window.letsPlotCallQueue = [];\n",
              "                        var div = document.createElement(\"div\");\n",
              "                        div.style.color = 'darkred';\n",
              "                        div.textContent = 'Error loading Lets-Plot JS';\n",
              "                        document.getElementById(\"lZRXqH\").appendChild(div);\n",
              "                    };\n",
              "                    var e = document.getElementById(\"lZRXqH\");\n",
              "                    e.appendChild(script);\n",
              "                })()\n",
              "            </script>\n",
              "            "
            ]
          },
          "metadata": {},
          "output_type": "display_data"
        }
      ],
      "source": [
        "import numpy as np\n",
        "import pandas as pd\n",
        "from sklearn.model_selection import train_test_split\n",
        "from sklearn.linear_model import LogisticRegression\n",
        "from sklearn.linear_model import Lasso\n",
        "from sklearn.metrics import precision_score, recall_score, f1_score, root_mean_squared_error\n",
        "import statsmodels.api as sm\n",
        "from lets_plot import *\n",
        "LetsPlot.setup_html()"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "## Before we do anything more\n",
        "\n",
        "Please create a data folder called `data` to store all the different data sets in this course if you haven't already done so."
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "# 📋 Lab Tasks\n",
        "\n",
        "## Part I - Cleaning a new data set `monitor` (30 min)\n",
        "\n",
        "The first step is to load the dataset. In this lab, we will be using the 2020 Wellcome Trust Global Monitor (`monitor`) dataset which contains data. It includes variables associated with trust in various actors and demographic information:\n",
        "\n",
        "-   `science_decreases_jobs`: binary outcome where respondents are asked whether they believe science and technology will decrease (versus increase, or have no impact on) the amount of jobs available. \n",
        "-   `trust_neighbours`: Trust People in Neighbourhood\n",
        "-   `trust_government`: Trust the National Government in This Country\n",
        "-   `trust_scientists`: Trust Scientists in This Country\n",
        "-   `trust_journalists`: Trust Journalists in This Country\n",
        "-   `trust_doctors`: Trust Doctors and Nurses in This Country\n",
        "-   `trust_ngos`: Trust People Who Work at Charitable Organizations or NGOs in This Country\n",
        "-   `trust_trad_healers`: Trust Traditional Healers in This Country\n",
        "-   `age`: Age (years)\n",
        "-   `gender`: Male / Female\n",
        "-   `education`: Education Level (Elementary / Secondary / Tertiary)\n",
        "-   `hh_income`: Per Capita Income Quintiles\n",
        "-   `subj_income`: Feelings About Household Income\n",
        "-   `emp_status`: Employment Status\n",
        "\n",
        "👉 **NOTE:** With variables such as `trust_neighbours`, we recoded the original question to a trust / no trust binary mainly for the sake of simplicity. This simplification can be useful as it reduces the number of parameters in our model. However, there may be distinct differences between each level that may produce different results."
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 11,
      "metadata": {},
      "outputs": [],
      "source": [
        "monitor = pd.read_csv(\"data/wellcome_global_monitor.csv\")"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "📝**Task:** Check the dimensions of the dataframe."
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 4,
      "metadata": {},
      "outputs": [
        {
          "name": "stdout",
          "output_type": "stream",
          "text": [
            "(119088, 14)\n"
          ]
        }
      ],
      "source": [
        "# Code here"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "📝**Task:** Here's another check we can perform quickly in `pandas`.  If we use the `isnull` method and then calculate the mean, we can check the proportion of missing values contained within each variable."
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 7,
      "metadata": {},
      "outputs": [],
      "source": [
        "# Code here"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "📝**Task:** We need to perform some variable transformations. Normalise `age`, convert all logical variables to integer variables (check out `astype`), and code all categorical features as one-hot encoded dummies."
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 6,
      "metadata": {},
      "outputs": [],
      "source": [
        "# Code here"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "📝**Task:** How many observations fall into each class of the variable `science_decreases_jobs`? Comment. What are the challenges you would expect when training a model?"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 2,
      "metadata": {},
      "outputs": [],
      "source": [
        "# Code here"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "📝**Task:** Come up with one or two summary visualisations using either `monitor` or `monitor_cleaned`!"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 1,
      "metadata": {},
      "outputs": [],
      "source": [
        "# Code here"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "## Part II - Fitting a logistic regression (40 min)\n",
        "\n",
        "We want to perform a logistic regression using the available variables in the dataset to predict whether an individual thinks that science will decrease jobs.\n",
        "\n",
        "👨🏻‍🏫 **TEACHING MOMENT:** Your class teacher will formalize the logistic regression in the context of the data at our disposal.\n",
        "\n",
        "Now, we need to split the data into training and testing sets. Having a test set will help us evaluate how well the model performs.\n",
        "\n",
        "-   In the training phase, we will use part of the data to fit the logistic regression model.\n",
        "-   In the testing phase, we will assess the model's performance on the remaining data (test set) which was not used during training."
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 40,
      "metadata": {},
      "outputs": [],
      "source": [
        "# Divide data into features (X) and targets (y) \n",
        "X = monitor_cleaned.drop(\"science_decreases_jobs\", axis=1).astype(float)\n",
        "X = sm.add_constant(X)\n",
        "y = monitor_cleaned[\"science_decreases_jobs\"].astype(int)\n",
        "\n",
        "# Perform a 0.75 / 0.25 train / test\n",
        "X_train, X_test, y_train, y_test = train_test_split(X, y, stratify=y, random_state=123)"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "💡**Tip:** When using `train_test_split` try specifying `stratify = y`. This ensures that the training and test sets will have identical proportions of the outcome.\n",
        "\n",
        "We will now fit a logistic regression model on the training data using `statsmodels`."
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 41,
      "metadata": {},
      "outputs": [
        {
          "name": "stdout",
          "output_type": "stream",
          "text": [
            "Optimization terminated successfully.\n",
            "         Current function value: 0.595142\n",
            "         Iterations 5\n",
            "                             Logit Regression Results                             \n",
            "==================================================================================\n",
            "Dep. Variable:     science_decreases_jobs   No. Observations:                89316\n",
            "Model:                              Logit   Df Residuals:                    89292\n",
            "Method:                               MLE   Df Model:                           23\n",
            "Date:                    Thu, 06 Feb 2025   Pseudo R-squ.:                 0.02893\n",
            "Time:                            09:32:51   Log-Likelihood:                -53156.\n",
            "converged:                           True   LL-Null:                       -54739.\n",
            "Covariance Type:                nonrobust   LLR p-value:                     0.000\n",
            "===========================================================================================================================\n",
            "                                                              coef    std err          z      P>|z|      [0.025      0.975]\n",
            "---------------------------------------------------------------------------------------------------------------------------\n",
            "const                                                      -0.2195      0.039     -5.562      0.000      -0.297      -0.142\n",
            "trust_neighbours                                           -0.0367      0.018     -2.009      0.045      -0.073      -0.001\n",
            "trust_government                                           -0.4455      0.016    -27.654      0.000      -0.477      -0.414\n",
            "trust_scientists                                           -0.1670      0.021     -8.057      0.000      -0.208      -0.126\n",
            "trust_journalists                                          -0.2412      0.016    -14.794      0.000      -0.273      -0.209\n",
            "trust_doctors                                              -0.1147      0.022     -5.187      0.000      -0.158      -0.071\n",
            "trust_ngos                                                 -0.1490      0.017     -8.543      0.000      -0.183      -0.115\n",
            "trust_trad_healers                                         -0.0869      0.015     -5.654      0.000      -0.117      -0.057\n",
            "age                                                         0.1614      0.008     20.614      0.000       0.146       0.177\n",
            "gender_Male                                                -0.0193      0.015     -1.267      0.205      -0.049       0.011\n",
            "education_Secondary (8-15 years)                            0.3059      0.024     12.690      0.000       0.259       0.353\n",
            "education_Tertiary                                          0.2953      0.027     11.105      0.000       0.243       0.347\n",
            "hh_income_Middle 20%                                        0.0646      0.023      2.762      0.006       0.019       0.110\n",
            "hh_income_Poorest 20%                                       0.0757      0.026      2.954      0.003       0.025       0.126\n",
            "hh_income_Richest 20%                                      -0.0322      0.022     -1.490      0.136      -0.075       0.010\n",
            "hh_income_Second 20%                                        0.0479      0.025      1.953      0.051      -0.000       0.096\n",
            "subj_income_Finding it very difficult on present income     0.1362      0.027      5.052      0.000       0.083       0.189\n",
            "subj_income_Getting by on present income                   -0.0414      0.021     -2.013      0.044      -0.082      -0.001\n",
            "subj_income_Living comfortable on present income           -0.1643      0.024     -6.957      0.000      -0.211      -0.118\n",
            "emp_status_Employed full time for self                     -0.1293      0.025     -5.274      0.000      -0.177      -0.081\n",
            "emp_status_Employed part time do not want full time        -0.0133      0.033     -0.408      0.684      -0.077       0.051\n",
            "emp_status_Employed part time want full time               -0.0833      0.029     -2.859      0.004      -0.140      -0.026\n",
            "emp_status_Out of workforce                                -0.0545      0.020     -2.786      0.005      -0.093      -0.016\n",
            "emp_status_Unemployed                                       0.0275      0.031      0.878      0.380      -0.034       0.089\n",
            "===========================================================================================================================\n"
          ]
        }
      ],
      "source": [
        "logit_model = sm.Logit(y_train, X_train).fit()\n",
        "print(logit_model.summary())"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "👨🏻‍🏫 **TEACHING MOMENT:** Your class teacher will explain how to interpret the summary of the model ($p-$values, AIC ...).\n",
        "\n",
        "We can run the same model using `scikit-learn` by specifying the following code."
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 42,
      "metadata": {},
      "outputs": [
        {
          "data": {
            "text/html": [
              "<style>#sk-container-id-3 {\n",
              "  /* Definition of color scheme common for light and dark mode */\n",
              "  --sklearn-color-text: #000;\n",
              "  --sklearn-color-text-muted: #666;\n",
              "  --sklearn-color-line: gray;\n",
              "  /* Definition of color scheme for unfitted estimators */\n",
              "  --sklearn-color-unfitted-level-0: #fff5e6;\n",
              "  --sklearn-color-unfitted-level-1: #f6e4d2;\n",
              "  --sklearn-color-unfitted-level-2: #ffe0b3;\n",
              "  --sklearn-color-unfitted-level-3: chocolate;\n",
              "  /* Definition of color scheme for fitted estimators */\n",
              "  --sklearn-color-fitted-level-0: #f0f8ff;\n",
              "  --sklearn-color-fitted-level-1: #d4ebff;\n",
              "  --sklearn-color-fitted-level-2: #b3dbfd;\n",
              "  --sklearn-color-fitted-level-3: cornflowerblue;\n",
              "\n",
              "  /* Specific color for light theme */\n",
              "  --sklearn-color-text-on-default-background: var(--sg-text-color, var(--theme-code-foreground, var(--jp-content-font-color1, black)));\n",
              "  --sklearn-color-background: var(--sg-background-color, var(--theme-background, var(--jp-layout-color0, white)));\n",
              "  --sklearn-color-border-box: var(--sg-text-color, var(--theme-code-foreground, var(--jp-content-font-color1, black)));\n",
              "  --sklearn-color-icon: #696969;\n",
              "\n",
              "  @media (prefers-color-scheme: dark) {\n",
              "    /* Redefinition of color scheme for dark theme */\n",
              "    --sklearn-color-text-on-default-background: var(--sg-text-color, var(--theme-code-foreground, var(--jp-content-font-color1, white)));\n",
              "    --sklearn-color-background: var(--sg-background-color, var(--theme-background, var(--jp-layout-color0, #111)));\n",
              "    --sklearn-color-border-box: var(--sg-text-color, var(--theme-code-foreground, var(--jp-content-font-color1, white)));\n",
              "    --sklearn-color-icon: #878787;\n",
              "  }\n",
              "}\n",
              "\n",
              "#sk-container-id-3 {\n",
              "  color: var(--sklearn-color-text);\n",
              "}\n",
              "\n",
              "#sk-container-id-3 pre {\n",
              "  padding: 0;\n",
              "}\n",
              "\n",
              "#sk-container-id-3 input.sk-hidden--visually {\n",
              "  border: 0;\n",
              "  clip: rect(1px 1px 1px 1px);\n",
              "  clip: rect(1px, 1px, 1px, 1px);\n",
              "  height: 1px;\n",
              "  margin: -1px;\n",
              "  overflow: hidden;\n",
              "  padding: 0;\n",
              "  position: absolute;\n",
              "  width: 1px;\n",
              "}\n",
              "\n",
              "#sk-container-id-3 div.sk-dashed-wrapped {\n",
              "  border: 1px dashed var(--sklearn-color-line);\n",
              "  margin: 0 0.4em 0.5em 0.4em;\n",
              "  box-sizing: border-box;\n",
              "  padding-bottom: 0.4em;\n",
              "  background-color: var(--sklearn-color-background);\n",
              "}\n",
              "\n",
              "#sk-container-id-3 div.sk-container {\n",
              "  /* jupyter's `normalize.less` sets `[hidden] { display: none; }`\n",
              "     but bootstrap.min.css set `[hidden] { display: none !important; }`\n",
              "     so we also need the `!important` here to be able to override the\n",
              "     default hidden behavior on the sphinx rendered scikit-learn.org.\n",
              "     See: https://github.com/scikit-learn/scikit-learn/issues/21755 */\n",
              "  display: inline-block !important;\n",
              "  position: relative;\n",
              "}\n",
              "\n",
              "#sk-container-id-3 div.sk-text-repr-fallback {\n",
              "  display: none;\n",
              "}\n",
              "\n",
              "div.sk-parallel-item,\n",
              "div.sk-serial,\n",
              "div.sk-item {\n",
              "  /* draw centered vertical line to link estimators */\n",
              "  background-image: linear-gradient(var(--sklearn-color-text-on-default-background), var(--sklearn-color-text-on-default-background));\n",
              "  background-size: 2px 100%;\n",
              "  background-repeat: no-repeat;\n",
              "  background-position: center center;\n",
              "}\n",
              "\n",
              "/* Parallel-specific style estimator block */\n",
              "\n",
              "#sk-container-id-3 div.sk-parallel-item::after {\n",
              "  content: \"\";\n",
              "  width: 100%;\n",
              "  border-bottom: 2px solid var(--sklearn-color-text-on-default-background);\n",
              "  flex-grow: 1;\n",
              "}\n",
              "\n",
              "#sk-container-id-3 div.sk-parallel {\n",
              "  display: flex;\n",
              "  align-items: stretch;\n",
              "  justify-content: center;\n",
              "  background-color: var(--sklearn-color-background);\n",
              "  position: relative;\n",
              "}\n",
              "\n",
              "#sk-container-id-3 div.sk-parallel-item {\n",
              "  display: flex;\n",
              "  flex-direction: column;\n",
              "}\n",
              "\n",
              "#sk-container-id-3 div.sk-parallel-item:first-child::after {\n",
              "  align-self: flex-end;\n",
              "  width: 50%;\n",
              "}\n",
              "\n",
              "#sk-container-id-3 div.sk-parallel-item:last-child::after {\n",
              "  align-self: flex-start;\n",
              "  width: 50%;\n",
              "}\n",
              "\n",
              "#sk-container-id-3 div.sk-parallel-item:only-child::after {\n",
              "  width: 0;\n",
              "}\n",
              "\n",
              "/* Serial-specific style estimator block */\n",
              "\n",
              "#sk-container-id-3 div.sk-serial {\n",
              "  display: flex;\n",
              "  flex-direction: column;\n",
              "  align-items: center;\n",
              "  background-color: var(--sklearn-color-background);\n",
              "  padding-right: 1em;\n",
              "  padding-left: 1em;\n",
              "}\n",
              "\n",
              "\n",
              "/* Toggleable style: style used for estimator/Pipeline/ColumnTransformer box that is\n",
              "clickable and can be expanded/collapsed.\n",
              "- Pipeline and ColumnTransformer use this feature and define the default style\n",
              "- Estimators will overwrite some part of the style using the `sk-estimator` class\n",
              "*/\n",
              "\n",
              "/* Pipeline and ColumnTransformer style (default) */\n",
              "\n",
              "#sk-container-id-3 div.sk-toggleable {\n",
              "  /* Default theme specific background. It is overwritten whether we have a\n",
              "  specific estimator or a Pipeline/ColumnTransformer */\n",
              "  background-color: var(--sklearn-color-background);\n",
              "}\n",
              "\n",
              "/* Toggleable label */\n",
              "#sk-container-id-3 label.sk-toggleable__label {\n",
              "  cursor: pointer;\n",
              "  display: flex;\n",
              "  width: 100%;\n",
              "  margin-bottom: 0;\n",
              "  padding: 0.5em;\n",
              "  box-sizing: border-box;\n",
              "  text-align: center;\n",
              "  align-items: start;\n",
              "  justify-content: space-between;\n",
              "  gap: 0.5em;\n",
              "}\n",
              "\n",
              "#sk-container-id-3 label.sk-toggleable__label .caption {\n",
              "  font-size: 0.6rem;\n",
              "  font-weight: lighter;\n",
              "  color: var(--sklearn-color-text-muted);\n",
              "}\n",
              "\n",
              "#sk-container-id-3 label.sk-toggleable__label-arrow:before {\n",
              "  /* Arrow on the left of the label */\n",
              "  content: \"▸\";\n",
              "  float: left;\n",
              "  margin-right: 0.25em;\n",
              "  color: var(--sklearn-color-icon);\n",
              "}\n",
              "\n",
              "#sk-container-id-3 label.sk-toggleable__label-arrow:hover:before {\n",
              "  color: var(--sklearn-color-text);\n",
              "}\n",
              "\n",
              "/* Toggleable content - dropdown */\n",
              "\n",
              "#sk-container-id-3 div.sk-toggleable__content {\n",
              "  max-height: 0;\n",
              "  max-width: 0;\n",
              "  overflow: hidden;\n",
              "  text-align: left;\n",
              "  /* unfitted */\n",
              "  background-color: var(--sklearn-color-unfitted-level-0);\n",
              "}\n",
              "\n",
              "#sk-container-id-3 div.sk-toggleable__content.fitted {\n",
              "  /* fitted */\n",
              "  background-color: var(--sklearn-color-fitted-level-0);\n",
              "}\n",
              "\n",
              "#sk-container-id-3 div.sk-toggleable__content pre {\n",
              "  margin: 0.2em;\n",
              "  border-radius: 0.25em;\n",
              "  color: var(--sklearn-color-text);\n",
              "  /* unfitted */\n",
              "  background-color: var(--sklearn-color-unfitted-level-0);\n",
              "}\n",
              "\n",
              "#sk-container-id-3 div.sk-toggleable__content.fitted pre {\n",
              "  /* unfitted */\n",
              "  background-color: var(--sklearn-color-fitted-level-0);\n",
              "}\n",
              "\n",
              "#sk-container-id-3 input.sk-toggleable__control:checked~div.sk-toggleable__content {\n",
              "  /* Expand drop-down */\n",
              "  max-height: 200px;\n",
              "  max-width: 100%;\n",
              "  overflow: auto;\n",
              "}\n",
              "\n",
              "#sk-container-id-3 input.sk-toggleable__control:checked~label.sk-toggleable__label-arrow:before {\n",
              "  content: \"▾\";\n",
              "}\n",
              "\n",
              "/* Pipeline/ColumnTransformer-specific style */\n",
              "\n",
              "#sk-container-id-3 div.sk-label input.sk-toggleable__control:checked~label.sk-toggleable__label {\n",
              "  color: var(--sklearn-color-text);\n",
              "  background-color: var(--sklearn-color-unfitted-level-2);\n",
              "}\n",
              "\n",
              "#sk-container-id-3 div.sk-label.fitted input.sk-toggleable__control:checked~label.sk-toggleable__label {\n",
              "  background-color: var(--sklearn-color-fitted-level-2);\n",
              "}\n",
              "\n",
              "/* Estimator-specific style */\n",
              "\n",
              "/* Colorize estimator box */\n",
              "#sk-container-id-3 div.sk-estimator input.sk-toggleable__control:checked~label.sk-toggleable__label {\n",
              "  /* unfitted */\n",
              "  background-color: var(--sklearn-color-unfitted-level-2);\n",
              "}\n",
              "\n",
              "#sk-container-id-3 div.sk-estimator.fitted input.sk-toggleable__control:checked~label.sk-toggleable__label {\n",
              "  /* fitted */\n",
              "  background-color: var(--sklearn-color-fitted-level-2);\n",
              "}\n",
              "\n",
              "#sk-container-id-3 div.sk-label label.sk-toggleable__label,\n",
              "#sk-container-id-3 div.sk-label label {\n",
              "  /* The background is the default theme color */\n",
              "  color: var(--sklearn-color-text-on-default-background);\n",
              "}\n",
              "\n",
              "/* On hover, darken the color of the background */\n",
              "#sk-container-id-3 div.sk-label:hover label.sk-toggleable__label {\n",
              "  color: var(--sklearn-color-text);\n",
              "  background-color: var(--sklearn-color-unfitted-level-2);\n",
              "}\n",
              "\n",
              "/* Label box, darken color on hover, fitted */\n",
              "#sk-container-id-3 div.sk-label.fitted:hover label.sk-toggleable__label.fitted {\n",
              "  color: var(--sklearn-color-text);\n",
              "  background-color: var(--sklearn-color-fitted-level-2);\n",
              "}\n",
              "\n",
              "/* Estimator label */\n",
              "\n",
              "#sk-container-id-3 div.sk-label label {\n",
              "  font-family: monospace;\n",
              "  font-weight: bold;\n",
              "  display: inline-block;\n",
              "  line-height: 1.2em;\n",
              "}\n",
              "\n",
              "#sk-container-id-3 div.sk-label-container {\n",
              "  text-align: center;\n",
              "}\n",
              "\n",
              "/* Estimator-specific */\n",
              "#sk-container-id-3 div.sk-estimator {\n",
              "  font-family: monospace;\n",
              "  border: 1px dotted var(--sklearn-color-border-box);\n",
              "  border-radius: 0.25em;\n",
              "  box-sizing: border-box;\n",
              "  margin-bottom: 0.5em;\n",
              "  /* unfitted */\n",
              "  background-color: var(--sklearn-color-unfitted-level-0);\n",
              "}\n",
              "\n",
              "#sk-container-id-3 div.sk-estimator.fitted {\n",
              "  /* fitted */\n",
              "  background-color: var(--sklearn-color-fitted-level-0);\n",
              "}\n",
              "\n",
              "/* on hover */\n",
              "#sk-container-id-3 div.sk-estimator:hover {\n",
              "  /* unfitted */\n",
              "  background-color: var(--sklearn-color-unfitted-level-2);\n",
              "}\n",
              "\n",
              "#sk-container-id-3 div.sk-estimator.fitted:hover {\n",
              "  /* fitted */\n",
              "  background-color: var(--sklearn-color-fitted-level-2);\n",
              "}\n",
              "\n",
              "/* Specification for estimator info (e.g. \"i\" and \"?\") */\n",
              "\n",
              "/* Common style for \"i\" and \"?\" */\n",
              "\n",
              ".sk-estimator-doc-link,\n",
              "a:link.sk-estimator-doc-link,\n",
              "a:visited.sk-estimator-doc-link {\n",
              "  float: right;\n",
              "  font-size: smaller;\n",
              "  line-height: 1em;\n",
              "  font-family: monospace;\n",
              "  background-color: var(--sklearn-color-background);\n",
              "  border-radius: 1em;\n",
              "  height: 1em;\n",
              "  width: 1em;\n",
              "  text-decoration: none !important;\n",
              "  margin-left: 0.5em;\n",
              "  text-align: center;\n",
              "  /* unfitted */\n",
              "  border: var(--sklearn-color-unfitted-level-1) 1pt solid;\n",
              "  color: var(--sklearn-color-unfitted-level-1);\n",
              "}\n",
              "\n",
              ".sk-estimator-doc-link.fitted,\n",
              "a:link.sk-estimator-doc-link.fitted,\n",
              "a:visited.sk-estimator-doc-link.fitted {\n",
              "  /* fitted */\n",
              "  border: var(--sklearn-color-fitted-level-1) 1pt solid;\n",
              "  color: var(--sklearn-color-fitted-level-1);\n",
              "}\n",
              "\n",
              "/* On hover */\n",
              "div.sk-estimator:hover .sk-estimator-doc-link:hover,\n",
              ".sk-estimator-doc-link:hover,\n",
              "div.sk-label-container:hover .sk-estimator-doc-link:hover,\n",
              ".sk-estimator-doc-link:hover {\n",
              "  /* unfitted */\n",
              "  background-color: var(--sklearn-color-unfitted-level-3);\n",
              "  color: var(--sklearn-color-background);\n",
              "  text-decoration: none;\n",
              "}\n",
              "\n",
              "div.sk-estimator.fitted:hover .sk-estimator-doc-link.fitted:hover,\n",
              ".sk-estimator-doc-link.fitted:hover,\n",
              "div.sk-label-container:hover .sk-estimator-doc-link.fitted:hover,\n",
              ".sk-estimator-doc-link.fitted:hover {\n",
              "  /* fitted */\n",
              "  background-color: var(--sklearn-color-fitted-level-3);\n",
              "  color: var(--sklearn-color-background);\n",
              "  text-decoration: none;\n",
              "}\n",
              "\n",
              "/* Span, style for the box shown on hovering the info icon */\n",
              ".sk-estimator-doc-link span {\n",
              "  display: none;\n",
              "  z-index: 9999;\n",
              "  position: relative;\n",
              "  font-weight: normal;\n",
              "  right: .2ex;\n",
              "  padding: .5ex;\n",
              "  margin: .5ex;\n",
              "  width: min-content;\n",
              "  min-width: 20ex;\n",
              "  max-width: 50ex;\n",
              "  color: var(--sklearn-color-text);\n",
              "  box-shadow: 2pt 2pt 4pt #999;\n",
              "  /* unfitted */\n",
              "  background: var(--sklearn-color-unfitted-level-0);\n",
              "  border: .5pt solid var(--sklearn-color-unfitted-level-3);\n",
              "}\n",
              "\n",
              ".sk-estimator-doc-link.fitted span {\n",
              "  /* fitted */\n",
              "  background: var(--sklearn-color-fitted-level-0);\n",
              "  border: var(--sklearn-color-fitted-level-3);\n",
              "}\n",
              "\n",
              ".sk-estimator-doc-link:hover span {\n",
              "  display: block;\n",
              "}\n",
              "\n",
              "/* \"?\"-specific style due to the `<a>` HTML tag */\n",
              "\n",
              "#sk-container-id-3 a.estimator_doc_link {\n",
              "  float: right;\n",
              "  font-size: 1rem;\n",
              "  line-height: 1em;\n",
              "  font-family: monospace;\n",
              "  background-color: var(--sklearn-color-background);\n",
              "  border-radius: 1rem;\n",
              "  height: 1rem;\n",
              "  width: 1rem;\n",
              "  text-decoration: none;\n",
              "  /* unfitted */\n",
              "  color: var(--sklearn-color-unfitted-level-1);\n",
              "  border: var(--sklearn-color-unfitted-level-1) 1pt solid;\n",
              "}\n",
              "\n",
              "#sk-container-id-3 a.estimator_doc_link.fitted {\n",
              "  /* fitted */\n",
              "  border: var(--sklearn-color-fitted-level-1) 1pt solid;\n",
              "  color: var(--sklearn-color-fitted-level-1);\n",
              "}\n",
              "\n",
              "/* On hover */\n",
              "#sk-container-id-3 a.estimator_doc_link:hover {\n",
              "  /* unfitted */\n",
              "  background-color: var(--sklearn-color-unfitted-level-3);\n",
              "  color: var(--sklearn-color-background);\n",
              "  text-decoration: none;\n",
              "}\n",
              "\n",
              "#sk-container-id-3 a.estimator_doc_link.fitted:hover {\n",
              "  /* fitted */\n",
              "  background-color: var(--sklearn-color-fitted-level-3);\n",
              "}\n",
              "</style><div id=\"sk-container-id-3\" class=\"sk-top-container\"><div class=\"sk-text-repr-fallback\"><pre>LogisticRegression()</pre><b>In a Jupyter environment, please rerun this cell to show the HTML representation or trust the notebook. <br />On GitHub, the HTML representation is unable to render, please try loading this page with nbviewer.org.</b></div><div class=\"sk-container\" hidden><div class=\"sk-item\"><div class=\"sk-estimator fitted sk-toggleable\"><input class=\"sk-toggleable__control sk-hidden--visually\" id=\"sk-estimator-id-3\" type=\"checkbox\" checked><label for=\"sk-estimator-id-3\" class=\"sk-toggleable__label fitted sk-toggleable__label-arrow\"><div><div>LogisticRegression</div></div><div><a class=\"sk-estimator-doc-link fitted\" rel=\"noreferrer\" target=\"_blank\" href=\"https://scikit-learn.org/1.6/modules/generated/sklearn.linear_model.LogisticRegression.html\">?<span>Documentation for LogisticRegression</span></a><span class=\"sk-estimator-doc-link fitted\">i<span>Fitted</span></span></div></label><div class=\"sk-toggleable__content fitted\"><pre>LogisticRegression()</pre></div> </div></div></div></div>"
            ],
            "text/plain": [
              "LogisticRegression()"
            ]
          },
          "execution_count": 42,
          "metadata": {},
          "output_type": "execute_result"
        }
      ],
      "source": [
        "# Instantiate / fit a logistic regression model\n",
        "logit_model = LogisticRegression() \n",
        "logit_model.fit(X_train, y_train)"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "Next, we will generate predicted probabilities on both the training and testing sets from our logistic regression model using `predict_proba`."
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 43,
      "metadata": {},
      "outputs": [],
      "source": [
        "# Generate predictions on the training set\n",
        "train_probas = logit_model.predict_proba(X_train)[:,1]\n",
        "\n",
        "# Generate predictions on the testing set\n",
        "test_probas = logit_model.predict_proba(X_test)[:,1]"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "💡**Tip:** `predict_proba` produces a 2-d numpy array which predicts the probability of the outcome not occurring (the 1st dimension) and occurring (the 2nd dimension). To select the 2nd dimension (and all the rows), employ this indexing `[:,1]` when using this function."
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 44,
      "metadata": {},
      "outputs": [
        {
          "data": {
            "text/plain": [
              "array([0.26031154, 0.35727596, 0.46766817, 0.33238068, 0.22080559,\n",
              "       0.29925601, 0.28789836, 0.42926639, 0.36675032, 0.32610668])"
            ]
          },
          "execution_count": 44,
          "metadata": {},
          "output_type": "execute_result"
        }
      ],
      "source": [
        "test_probas[0:10]"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "We are going to set an arbitrary threshold $\\alpha$. All the scores that are higher than $\\alpha=0.6$ will be classified as 1 and the scores lower than $\\alpha=0.6$ will be classified as 0."
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 45,
      "metadata": {},
      "outputs": [],
      "source": [
        "# Convert predicted scores to binary outcomes for the training set\n",
        "train_preds = np.where(train_probas > 0.6, 1, 0)\n",
        "\n",
        "# Convert predicted scores to binary outcomes for the testing set\n",
        "test_preds = np.where(test_probas > 0.6, 1, 0)"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "👨🏻‍🏫 **TEACHING MOMENT:** Your class teacher will explain what precision and recall are."
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "Now, let's print the precision and recall for both training and testing sets."
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 59,
      "metadata": {},
      "outputs": [
        {
          "name": "stdout",
          "output_type": "stream",
          "text": [
            "Training set precision: 0.477\n",
            "Testing set precision: 0.467\n",
            "Training set recall: 0.001\n",
            "Testing set recall: 0.001\n"
          ]
        }
      ],
      "source": [
        "print(f\"Training set precision: {np.round(precision_score(y_train, train_preds), 3)}\")\n",
        "print(f\"Testing set precision: {np.round(precision_score(y_test, test_preds), 3)}\")\n",
        "print(f\"Training set recall: {np.round(recall_score(y_train, train_preds), 3)}\")\n",
        "print(f\"Testing set recall: {np.round(recall_score(y_test, test_preds), 3)}\")"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "❓**Question:** Compare the results you obtain for both the training and testing sets. Was this expected ? Does the model overfit ? underfit ?\n",
        "\n",
        "❓**Question:** What is the problem with setting an arbitrary threshold $\\alpha=0.6$ ? How should we expect the precision and the recall to behave if we increase the threshold $\\alpha$ ? If we decrease $\\alpha$ ?\n",
        "\n",
        "We will now compute the precision and recall for the test set for different values of the threshold $\\alpha$. Compute them for $\\alpha \\in \\{0.1, 0.2, 0.3, 0.4, 0.5, 0.6 \\}$."
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 17,
      "metadata": {},
      "outputs": [
        {
          "data": {
            "text/html": [
              "<div>\n",
              "<style scoped>\n",
              "    .dataframe tbody tr th:only-of-type {\n",
              "        vertical-align: middle;\n",
              "    }\n",
              "\n",
              "    .dataframe tbody tr th {\n",
              "        vertical-align: top;\n",
              "    }\n",
              "\n",
              "    .dataframe thead th {\n",
              "        text-align: right;\n",
              "    }\n",
              "</style>\n",
              "<table border=\"1\" class=\"dataframe\">\n",
              "  <thead>\n",
              "    <tr style=\"text-align: right;\">\n",
              "      <th></th>\n",
              "      <th>threshold</th>\n",
              "      <th>Precision</th>\n",
              "      <th>Recall</th>\n",
              "    </tr>\n",
              "  </thead>\n",
              "  <tbody>\n",
              "    <tr>\n",
              "      <th>0</th>\n",
              "      <td>0.1</td>\n",
              "      <td>0.302387</td>\n",
              "      <td>1.000000</td>\n",
              "    </tr>\n",
              "    <tr>\n",
              "      <th>1</th>\n",
              "      <td>0.2</td>\n",
              "      <td>0.315991</td>\n",
              "      <td>0.949459</td>\n",
              "    </tr>\n",
              "    <tr>\n",
              "      <th>2</th>\n",
              "      <td>0.3</td>\n",
              "      <td>0.382039</td>\n",
              "      <td>0.562648</td>\n",
              "    </tr>\n",
              "    <tr>\n",
              "      <th>3</th>\n",
              "      <td>0.4</td>\n",
              "      <td>0.440926</td>\n",
              "      <td>0.226340</td>\n",
              "    </tr>\n",
              "    <tr>\n",
              "      <th>4</th>\n",
              "      <td>0.5</td>\n",
              "      <td>0.509091</td>\n",
              "      <td>0.034212</td>\n",
              "    </tr>\n",
              "    <tr>\n",
              "      <th>5</th>\n",
              "      <td>0.6</td>\n",
              "      <td>0.477273</td>\n",
              "      <td>0.000778</td>\n",
              "    </tr>\n",
              "  </tbody>\n",
              "</table>\n",
              "</div>"
            ],
            "text/plain": [
              "   threshold  Precision    Recall\n",
              "0        0.1   0.302387  1.000000\n",
              "1        0.2   0.315991  0.949459\n",
              "2        0.3   0.382039  0.562648\n",
              "3        0.4   0.440926  0.226340\n",
              "4        0.5   0.509091  0.034212\n",
              "5        0.6   0.477273  0.000778"
            ]
          },
          "execution_count": 17,
          "metadata": {},
          "output_type": "execute_result"
        }
      ],
      "source": [
        "th_list = np.arange(0.1, 0.7, 0.1)\n",
        "\n",
        "def test_proba_th(th, prob_est, truth):\n",
        "  class_preds = np.where(prob_est > th, 1, 0)\n",
        "  prec = precision_score(truth, class_preds)\n",
        "  sens = recall_score(truth, class_preds)\n",
        "  return prec, sens\n",
        "\n",
        "metrics = [test_proba_th(th,train_probas,y_train) for th in th_list]\n",
        "\n",
        "output = pd.DataFrame({\"threshold\": th_list, \n",
        "                       \"Precision\": [m[0] for m in metrics],\n",
        "                       \"Recall\": [m[1] for m in metrics]})\n",
        "\n",
        "output"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "👨🏻‍🏫 **TEACHING MOMENT:** Your teacher will explain what the code chunk is doing.\n",
        "\n",
        "-   We specify a probability threshold range using `np.arange`. Note, when we specify an interval of 0.1, we need to set the maximum value to 0.7 to get 0.6.\n",
        "-   We build a function using `np.where` that creates a binary variable depending on the probability threshold selected and then calculates precision / recall scores.\n",
        "-   We loop the function over the different thresholds using list comprehension.\n",
        "-   We create a data frame with the thresholds and precision and recall scores. While we have a list of thresholds, we need to use list comprehension to get the correct dimension of `metrics`, which is a 2-d numpy array.\n",
        "\n",
        "Let's create a plot that shows performance over different thresholds, using colour to distinguish between the different evaluation metrics."
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 18,
      "metadata": {},
      "outputs": [
        {
          "data": {
            "text/html": [
              "   <div id=\"ZGWuUm\"></div>\n",
              "   <script type=\"text/javascript\" data-lets-plot-script=\"plot\">\n",
              "   \n",
              "   (function() {\n",
              "   // ----------\n",
              "   \n",
              "   var containerDiv = document.getElementById(\"ZGWuUm\");\n",
              "   var observer = new ResizeObserver(function(entries) {\n",
              "       for (let entry of entries) {\n",
              "           var width = containerDiv.clientWidth\n",
              "           if (entry.contentBoxSize && width > 0) {\n",
              "           \n",
              "               // Render plot\n",
              "               if (observer) {\n",
              "                   observer.disconnect();\n",
              "                   observer = null;\n",
              "               }\n",
              "\n",
              "               var plotSpec={\n",
              "\"data\":{\n",
              "\"threshold\":[0.1,0.2,0.30000000000000004,0.4,0.5,0.6,0.1,0.2,0.30000000000000004,0.4,0.5,0.6],\n",
              "\"variable\":[\"Precision\",\"Precision\",\"Precision\",\"Precision\",\"Precision\",\"Precision\",\"Recall\",\"Recall\",\"Recall\",\"Recall\",\"Recall\",\"Recall\"],\n",
              "\"value\":[0.3023870303193157,0.31599117694175055,0.3820394207562349,0.44092613964223887,0.509090909090909,0.4772727272727273,1.0,0.9494594194312796,0.5626481042654028,0.22634034360189573,0.034212085308056875,7.775473933649289E-4]\n",
              "},\n",
              "\"mapping\":{\n",
              "\"x\":\"threshold\",\n",
              "\"y\":\"value\",\n",
              "\"colour\":\"variable\",\n",
              "\"linetype\":\"variable\"\n",
              "},\n",
              "\"data_meta\":{\n",
              "\"series_annotations\":[{\n",
              "\"type\":\"float\",\n",
              "\"column\":\"threshold\"\n",
              "},{\n",
              "\"type\":\"str\",\n",
              "\"column\":\"variable\"\n",
              "},{\n",
              "\"type\":\"float\",\n",
              "\"column\":\"value\"\n",
              "}]\n",
              "},\n",
              "\"theme\":{\n",
              "\"legend_position\":\"bottom\",\n",
              "\"panel_grid_minor\":{\n",
              "\"blank\":true\n",
              "},\n",
              "\"panel_grid_major_x\":{\n",
              "\"blank\":true\n",
              "}\n",
              "},\n",
              "\"guides\":{\n",
              "\"x\":{\n",
              "\"title\":\"Probability threshold\"\n",
              "},\n",
              "\"y\":{\n",
              "\"title\":\"Score\"\n",
              "},\n",
              "\"colour\":{\n",
              "\"title\":\"Metric\"\n",
              "},\n",
              "\"linetype\":{\n",
              "\"title\":\"Metric\"\n",
              "}\n",
              "},\n",
              "\"kind\":\"plot\",\n",
              "\"scales\":[],\n",
              "\"layers\":[{\n",
              "\"geom\":\"line\",\n",
              "\"mapping\":{\n",
              "},\n",
              "\"data_meta\":{\n",
              "},\n",
              "\"data\":{\n",
              "}\n",
              "}],\n",
              "\"metainfo_list\":[],\n",
              "\"spec_id\":\"1\"\n",
              "};\n",
              "               window.letsPlotCall(function() {\n",
              "       \n",
              "               var toolbar = null;\n",
              "               var plotContainer = containerDiv;               \n",
              "               \n",
              "                   var options = {\n",
              "                       sizing: {\n",
              "                           width_mode: \"min\",\n",
              "                           height_mode: \"scaled\",\n",
              "                           width: width\n",
              "                       }\n",
              "                   };\n",
              "                   var fig = LetsPlot.buildPlotFromProcessedSpecs(plotSpec, -1, -1, plotContainer, options);\n",
              "                   if (toolbar) {\n",
              "                     toolbar.bind(fig);\n",
              "                   }\n",
              "               });\n",
              "               \n",
              "               break;\n",
              "           }\n",
              "       }\n",
              "   });\n",
              "   \n",
              "   observer.observe(containerDiv);\n",
              "   \n",
              "   // ----------\n",
              "   })();\n",
              "   \n",
              "   </script>"
            ],
            "text/plain": [
              "<lets_plot.plot.core.PlotSpec at 0x1b13e8c7cb0>"
            ]
          },
          "execution_count": 18,
          "metadata": {},
          "output_type": "execute_result"
        }
      ],
      "source": [
        "output_melted = pd.melt(output, id_vars=\"threshold\")\n",
        "\n",
        "(\n",
        "  ggplot(output_melted, aes(\"threshold\", \"value\", colour = \"variable\", linetype = \"variable\")) +\n",
        "  geom_line() +\n",
        "  theme(panel_grid_minor = element_blank(),\n",
        "        panel_grid_major_x = element_blank(),\n",
        "        legend_position = \"bottom\") +\n",
        "  labs(x = \"Probability threshold\", y = \"Score\", colour = \"Metric\", linetype = \"Metric\")\n",
        ")"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "📝**Task:** Try redefining `class_metrics` to also include `f1_score` and see what happens."
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 3,
      "metadata": {},
      "outputs": [],
      "source": [
        "# Code here"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "📝**Task:** Based on the precision and recall values you calculated, create a precision-recall plot to visualize the relationship between these two metrics. How does the shape of the precision-recall curve help you assess model performance ? What trade-offs do you observe between precision and recall?"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 6,
      "metadata": {},
      "outputs": [],
      "source": [
        "# Code here"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "## Working with time-series cross-sectional data in machine learning (30 mins)\n",
        "\n",
        "Economists and political scientists constantly work with data that has a cross-sectional (e.g. between country) and time-series (e.g. within countries over years) structure.  This poses an interesting challenge for when we perform a train / test split. If we use `train_test_split` on such data, the function will assume that it is cross-sectional and perform an initial split without taking the time-series dimension into account.  In doing so, we run the risk of:\n",
        "\n",
        "-   Using future values to predict the past\n",
        "-   Using information from the same country in the training and testing sets\n",
        "\n",
        "Naturally, a few solutions follow from this issue:\n",
        "\n",
        "-   Use earlier years to train the model and later years to test it.\n",
        "-   Use some countries to train the model and other countries to test it.\n",
        "\n",
        "📝**Task:** Use GDP per capita in the Gapminder data set (see the Week 2 lab) to predict life expectancy.\n",
        "\n",
        "Open the data set."
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 209,
      "metadata": {},
      "outputs": [],
      "source": [
        "# Code here"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "Convert the variable names to snake case (see Lab 2)."
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 210,
      "metadata": {},
      "outputs": [],
      "source": [
        "# Code here"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "### Splitting the data by year\n",
        "\n",
        "Use data from 2007 as the testing data, and all other years as the training data."
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 211,
      "metadata": {},
      "outputs": [],
      "source": [
        "# Code here"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "Split the training / testing data into outcomes and features (remember to convert the series into a one-column data frame)."
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 212,
      "metadata": {},
      "outputs": [],
      "source": [
        "# Code here"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "Build a linear model on the training set, and evaluate it on the test set, using RMSE"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 214,
      "metadata": {},
      "outputs": [
        {
          "name": "stdout",
          "output_type": "stream",
          "text": [
            "On average, our model is approximately 10.05 years off!\n"
          ]
        }
      ],
      "source": [
        "# Code here"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "### Splitting the data by country\n",
        "\n",
        "Create an array of all unique country values."
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 215,
      "metadata": {},
      "outputs": [],
      "source": [
        "# Code here"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "Randomly sample 36 countries (~25% of all countries in `gapminder`) to be in the test set.\n",
        "\n",
        "💁**Tip:** You'll need a couple of functions from `numpy.random` first, to set a random seed (`seed`) for reproduceability and, second, to select the number of countries (`choice`). "
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 216,
      "metadata": {},
      "outputs": [],
      "source": [
        "# Code here"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "Split the data into training and testing sets."
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 217,
      "metadata": {},
      "outputs": [],
      "source": [
        "# Code here"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "Split the training / testing data into outcomes and features (remember to convert the series into a one-column data frame)."
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 5,
      "metadata": {},
      "outputs": [],
      "source": [
        "# Code here"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "Build a linear model on the training set, and evaluate it on the test set, using RMSE"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 4,
      "metadata": {},
      "outputs": [],
      "source": [
        "# Code here"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "👉 **NOTE:** When working with TSCS data, you may need to calculate 1 or more period lags for all features to reduce the possibility of reverse causation.  We have not opted to do this in order to focus on the train / test split itself but these should nevertheless be considered in the real world."
      ]
    }
  ],
  "metadata": {
    "kernelspec": {
      "display_name": "Python 3",
      "language": "python",
      "name": "python3"
    },
    "language_info": {
      "codemirror_mode": {
        "name": "ipython",
        "version": 3
      },
      "file_extension": ".py",
      "mimetype": "text/x-python",
      "name": "python",
      "nbconvert_exporter": "python",
      "pygments_lexer": "ipython3",
      "version": "3.12.3"
    }
  },
  "nbformat": 4,
  "nbformat_minor": 4
}
